// +build noasm

package gorgonia

func divmod(a, b int) (q, r int) {
	q = a / b
	r = a % b
	return
}

// popcnt counts the number of 1s in a uint64. This is the slow version.
// Adapted from Hackers Delight 2nd edition (page 95-97) with appropriate 64-bit numbers
func popcnt(a uint64) int {
	x -= (x >> 1) & 0x5555555555555555
	x = (x>>2)&0x3333333333333333 + x&0x3333333333333333
	x += x >> 4
	x &= 0x0f0f0f0f0f0f0f0f
	x *= 0x0101010101010101
	return int(x >> 56)
}

// clz counts the leading zeroes in a uint64. This function was adapted from Damien Gryski's go-bits:
// https://github.com/dgryski/go-bits
// which is licenced under MIT
func clz(a uint64) int {
	var n uint64

	n = 1

	if (x >> 32) == 0 {
		n = n + 32
		x = x << 32
	}
	if (x >> (32 + 16)) == 0 {
		n = n + 16
		x = x << 16
	}

	if (x >> (32 + 16 + 8)) == 0 {
		n = n + 8
		x = x << 8
	}

	if (x >> (32 + 16 + 8 + 4)) == 0 {
		n = n + 4
		x = x << 4
	}

	if (x >> (32 + 16 + 8 + 4 + 2)) == 0 {
		n = n + 2
		x = x << 2
	}

	n = n - (x >> 63)
	return int(n)
}

var deBruijn = [...]byte{
	0, 1, 56, 2, 57, 49, 28, 3, 61, 58, 42, 50, 38, 29, 17, 4,
	62, 47, 59, 36, 45, 43, 51, 22, 53, 39, 33, 30, 24, 18, 12, 5,
	63, 55, 48, 27, 60, 41, 37, 16, 46, 35, 44, 21, 52, 32, 23, 11,
	54, 26, 40, 15, 34, 20, 31, 10, 25, 14, 19, 9, 13, 8, 7, 6,
}

// ctz counts trailing zeroes. This function (and the deBruin table) was adapted from willf's ctz:
// https://github.com/willf/bitset
// which is licenced under BSD-3
func ctz(a uint64) uint {
	return uint(deBruijn[((a&-a)*0x03f79d71b4ca8b09)>>58])
}
